# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .base_traintpl import BaseTrainTPL
from .gd_traintpl import GDTrainTPL
from .general_traintpl import GeneralTrainTPL
from .atkt_traintpl import AtktTrainTPL
from .dcd_traintpl import DCDTrainTPL
from .adversarial_traintpl import AdversarialTrainTPL
from .group_cd_traintpl import GroupCDTrainTPL
